#include <stdlib.h>
#include <stdio.h>
#include "io.h"
#include "state.h"
#include "mc_string.h"
#include "string.h"

#define SPC 0x20
#define CR 0x0D
#define LF 0x0A
#define TAB 0x09

// codes for tags that mask other tags.
// for example </B> and <PL> mask <B>.
int mask1[3] = {11,4,0};							// <B>
int mask2[6] = {2,12,4,5,0,0};						// <EM>
int mask3[3] = {13,4,0};							// <I>
int mask5[3] = {15,4,0};							// <S>
int mask6[3] = {16,4,0};							// <TT>
int mask7[3] = {17,4,0};							// <U>
int mask20[9] = {20,21,22,23,24,25,26,27,0};		// <r..w>
int mask40[11] = {50,41,42,43,44,45,46,47,48,49,0};	// <0..9>

// copies null-terminated integer arrays
void arraycopy(int* target, int* source)
{
	int i=0;
	for(;;)
	{
		target[i]=source[i];
		if( target[i] == 0 )
			break;
		i++;
	}
}

// checks if the given tag is masked before hte next text segment
short masked(struct doc_node* cur)
{
	int mask[11];
	int i;
	int type = cur->type;

	if ( ( type == 2 ) && ( mask2[5] == 1 ) )
	{
		mask2[5]=0;
		return 1;
	}

	// compute masked tags...
	if ( type == 1 ) arraycopy(mask,mask1);
	else if ( type == 2) arraycopy(mask,mask2);
	else if ( type == 3) arraycopy(mask,mask3);
	else if ( type == 5) arraycopy(mask,mask5);
	else if ( type == 6) arraycopy(mask,mask6);
	else if ( type == 7) arraycopy(mask,mask7);
	else if ( ( type < 28) && ( type > 19 ) )
	{
		arraycopy(mask,mask20);
		mask[ type - 20 ] += 10;
	}
	else if ( ( type < 50) && ( type > 39 ) )
	{
		arraycopy(mask,mask40);
		mask[ type - 40 ] += 10;
	}

	cur = cur->next;

	// search for masked tag before text segment
	while(cur->type != 0)
	{
		i=0;
		while( mask[i] != 0 )
		{
			if ( cur->type == mask[i] )
			{
				if ( ( type == 2 ) && ( cur->type == 2 ) )
					mask2[5]=1;
				return 1;
			}
			i++;
		}

		cur = cur->next;
	}
	return 0;
}

/*
 * remove redundant tags and condense, simplify whitespace
 */
void first_pass(struct doc_node* doc_root)
{
	struct doc_node*   doc_cur    = doc_root;
	struct statestack* state_root = newstack();
	struct statestack* state_cur  = state_root;

	short space_to_borrow = 0;
	struct statestack saved_state = *state_root;
	struct doc_node* old_textnode = NULL;

	while ( doc_cur != NULL )
	{
		if ( is_text(doc_cur) )
		{
			long i;
			char* str = doc_cur->text->contents;
			short in_ws = 0;

			if ( space_to_borrow )
			{
				// space relocation, simplification
				if ( (str[0]==SPC) || (str[0]==CR) ||
					 (str[0]==LF) || (str[0]==TAB) )
				{
					if ( ( test_u(&saved_state) == test_u(state_cur) ) &&
						( get_size(&saved_state) == get_size(state_cur) ) &&
						( test_tt(&saved_state) == test_tt(state_cur) ))
					{
						if ( ( saved_state.u_level == 0 ) || 
							( saved_state.color == get_color(state_cur) ) )
						{
							// ok to move string
							struct mc_string* buffer = strnew();
							struct mc_string* old_str = old_textnode->text;
							char tmpch;

							for(;;)
							{
								tmpch = strtaildelc(old_str);
								if ( ( tmpch == SPC ) || ( tmpch == CR ) ||
									( tmpch == LF ) || ( tmpch == TAB ) ||
									( tmpch == '>' ) )
								{
									strheadaddc( buffer, tmpch );
								}
								else
								{
									if ( (tmpch==0) && 
											(old_textnode->prev!=NULL) )
										delete_tag(old_textnode);
									else
										strtailaddc(old_str,tmpch);
									break;
								}
							}
							strheadadd(doc_cur->text, buffer->contents);
							free(buffer);
						}
					}
				}
			}

			if ( !test_tt( state_cur ) )
			{
				// ok to compress spaces
				for(i=0; i<doc_cur->text->length; i++)
				{
					if ( str[i] == SPC )
					{
						if ( in_ws )
							str[i]='>';
						in_ws = 1;
					}
					else if ( str[i] == CR )
					{
						if ( in_ws )
							str[i] = '>';
						else
							str[i] = SPC;
						in_ws = 1;
					}
					else if ( str[i] == LF )
					{	
						if ( in_ws )
							str[i] = '>';
						else
							str[i] = SPC;
						in_ws = 1;
					}
					else if ( str[i] == TAB )
					{	
						if ( in_ws )
							str[i] = '>';
						else
							str[i] = SPC;
						in_ws = 1;
					}
					else if ( str[i] == '>' )
						in_ws = 1;
					else
						in_ws = 0;
				}
			}

			if (( str[doc_cur->text->length-1] == SPC ) ||
				  	( str[doc_cur->text->length-1] == '>' )	)
				space_to_borrow = 1;
			else
				space_to_borrow = 0;

			old_textnode = doc_cur;
			saved_state = *state_cur;
		}

		else if ( is_opentag(doc_cur) )
		{
			// update state
			state_cur = newstate(state_cur);
			if ( doc_cur->type == 1 ) // B
			{
				set_b( state_cur );
			}
			else if ( doc_cur->type == 2 ) // EM
			{
				toggle_em( state_cur );
			}
			else if ( doc_cur->type == 3 ) // I
			{
				set_i( state_cur );
			}
			else if ( doc_cur->type == 4 ) // PL
			{
				clear_u ( state_cur );
				clear_b ( state_cur );
				clear_em( state_cur );
				clear_i ( state_cur );
				clear_s ( state_cur );
				clear_tt( state_cur );
			}
			else if ( doc_cur->type == 5 ) // S
			{
				set_s( state_cur );
			}
			else if ( doc_cur->type == 6 ) // TT
			{
				set_tt( state_cur );
			}
			else if ( doc_cur->type == 7 ) // U
			{
				inc_u( state_cur );
			}
			else if ( (doc_cur->type>=20) && (doc_cur->type<=27) )
			{
				set_color( state_cur, doc_cur->type-20 );
			}
			else if ( (doc_cur->type>=40) && (doc_cur->type<=49) )
			{
				set_size( state_cur, doc_cur->type-40 );
			}

			// redundant if state did not change after its appearance
			if ( equivstates( state_cur, state_cur->up ) )
			{
				doc_cur = delete_tag(doc_cur);
				state_cur->tag_deleted = 1;
				continue;
			}
		}

		else // is closetag
		{
			// check if our matching opentag was redundant
			if ( state_cur->tag_deleted == 1 )
			{
				doc_cur = delete_tag(doc_cur);
				state_cur = restorestate(state_cur);
				continue;
			}
			state_cur = restorestate(state_cur);
		}
		doc_cur = doc_cur->next;
	}
}


/*
 * remove tags that are masked before any text is dislplayed
 */
short opt_mask(struct doc_node* doc_root)
{
	struct doc_node*   doc_cur    = doc_root;
	struct statestack* state_root = newstack();
	struct statestack* state_cur  = state_root;

	short em_flag=0;
	short change=0;

	while ( doc_cur != NULL )
	{
		if ( !is_text(doc_cur) )
		{
			if ( is_opentag(doc_cur) )
			{
				state_cur = newstate(state_cur);
				if( (em_flag==1) || (masked(doc_cur)==1) )
				{
					if (doc_cur->type == 2)
						em_flag = 1 - em_flag;
					doc_cur = delete_tag(doc_cur);
					state_cur->tag_deleted = 1;
					change++;
					continue;
				}
			}
			else // is closetag
			{
				if ( state_cur->tag_deleted == 1 )
				{
					doc_cur = delete_tag(doc_cur);
					state_cur = restorestate(state_cur);
					change++;
					continue;
				}
				// close immediatly followed by an open
				if ( doc_cur->type == doc_cur->next->type+10 )
				{
					doc_cur = delete_tag(doc_cur);
					doc_cur = delete_tag(doc_cur);
					change++;
					continue;
				}
					
				state_cur = restorestate(state_cur);
			}
		}
		doc_cur = doc_cur->next;
	}
	return change;
}

